(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var Accounts = Package['accounts-base'].Accounts;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var SandstormPermissions = Package['sandstorm-permissions'].SandstormPermissions;

/* Package-scope variables */
var SandstormBackend, inMeteor, promiseToFuture, waitPromise, SANDSTORM_ALTHOME;

(function () {

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                     //
// packages/sandstorm-backend/sandstorm-backend.js                                                                     //
//                                                                                                                     //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                       //
// Sandstorm - Personal Cloud Sandbox                                                                                  // 1
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                               // 2
// All rights reserved.                                                                                                // 3
//                                                                                                                     // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                                     // 5
// you may not use this file except in compliance with the License.                                                    // 6
// You may obtain a copy of the License at                                                                             // 7
//                                                                                                                     // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                                        // 9
//                                                                                                                     // 10
// Unless required by applicable law or agreed to in writing, software                                                 // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                                   // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                            // 13
// See the License for the specific language governing permissions and                                                 // 14
// limitations under the License.                                                                                      // 15
                                                                                                                       // 16
var Capnp = Npm.require("capnp");                                                                                      // 17
var Backend = Capnp.importSystem("sandstorm/backend.capnp").Backend;                                                   // 18
var Crypto = Npm.require("crypto");                                                                                    // 19
var Future = Npm.require("fibers/future");                                                                             // 20
var Promise = Npm.require("es6-promise").Promise;                                                                      // 21
                                                                                                                       // 22
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {                                                    // 23
  callback();                                                                                                          // 24
});                                                                                                                    // 25
                                                                                                                       // 26
inMeteor = function (callback) {                                                                                       // 27
  // Calls the callback in a Meteor context.  Returns a Promise for its result.                                        // 28
  return new Promise(function (resolve, reject) {                                                                      // 29
    inMeteorInternal(function () {                                                                                     // 30
      try {                                                                                                            // 31
        resolve(callback());                                                                                           // 32
      } catch (err) {                                                                                                  // 33
        reject(err);                                                                                                   // 34
      }                                                                                                                // 35
    });                                                                                                                // 36
  });                                                                                                                  // 37
}                                                                                                                      // 38
                                                                                                                       // 39
promiseToFuture = function (promise) {                                                                                 // 40
  var result = new Future();                                                                                           // 41
  promise.then(result.return.bind(result), result.throw.bind(result));                                                 // 42
  return result;                                                                                                       // 43
}                                                                                                                      // 44
                                                                                                                       // 45
waitPromise = function (promise) {                                                                                     // 46
  return promiseToFuture(promise).wait();                                                                              // 47
}                                                                                                                      // 48
                                                                                                                       // 49
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;                                                           // 50
                                                                                                                       // 51
SandstormBackend = function(db, backendCap) {                                                                          // 52
  this._db = db;                                                                                                       // 53
  this.runningGrains = {};                                                                                             // 54
  this._backendCap = backendCap                                                                                        // 55
}                                                                                                                      // 56
                                                                                                                       // 57
SandstormBackend.prototype.cap = function() {                                                                          // 58
  return this._backendCap;                                                                                             // 59
}                                                                                                                      // 60
                                                                                                                       // 61
SandstormBackend.prototype.shutdownGrain = function (grainId, ownerId, keepSessions) {                                 // 62
  if (!keepSessions) {                                                                                                 // 63
    Sessions.remove({grainId: grainId});                                                                               // 64
    delete this.runningGrains[grainId];                                                                                // 65
  }                                                                                                                    // 66
                                                                                                                       // 67
  var grain = this._backendCap.getGrain(ownerId, grainId).supervisor;                                                  // 68
  return grain.shutdown().then(function () {                                                                           // 69
    grain.close();                                                                                                     // 70
    throw new Error("expected shutdown() to throw disconnected");                                                      // 71
  }, function (err) {                                                                                                  // 72
    grain.close();                                                                                                     // 73
    if (err.kjType !== "disconnected") {                                                                               // 74
      throw err;                                                                                                       // 75
    }                                                                                                                  // 76
  });                                                                                                                  // 77
}                                                                                                                      // 78
                                                                                                                       // 79
SandstormBackend.prototype.deleteGrain = function (grainId, ownerId) {                                                 // 80
  // We leave it up to the caller if they want to actually wait, but some don't so we report                           // 81
  // exceptions.                                                                                                       // 82
  return this._backendCap.deleteGrain(ownerId, grainId).catch(function (err) {                                         // 83
    console.error("problem deleting grain " + grainId + ":", err.message);                                             // 84
    throw err;                                                                                                         // 85
  });                                                                                                                  // 86
}                                                                                                                      // 87
                                                                                                                       // 88
                                                                                                                       // 89
SandstormBackend.prototype.openGrain = function (grainId, isRetry) {                                                   // 90
  // Create a Cap'n Proto connection to the given grain. Note that this function does not actually                     // 91
  // verify that the connection succeeded. Instead, if an RPC call to the connection fails, check                      // 92
  // shouldRestartGrain(). If it returns true, call continueGrain() and then openGrain()                               // 93
  // again with isRetry = true, and then retry.                                                                        // 94
  //                                                                                                                   // 95
  // Must be called in a Meteor context.                                                                               // 96
                                                                                                                       // 97
  if (isRetry) {                                                                                                       // 98
    // Since this is a retry, try starting the grain even if we think it's already running.                            // 99
    return this.continueGrain(grainId);                                                                                // 100
  } else {                                                                                                             // 101
    // Start the grain if it is not running.                                                                           // 102
    var runningGrain = this.runningGrains[grainId];                                                                    // 103
    if (runningGrain) {                                                                                                // 104
      return waitPromise(runningGrain);                                                                                // 105
    } else {                                                                                                           // 106
      return this.continueGrain(grainId);                                                                              // 107
    }                                                                                                                  // 108
  }                                                                                                                    // 109
}                                                                                                                      // 110
                                                                                                                       // 111
SandstormBackend.shouldRestartGrain = function (error, retryCount) {                                                   // 112
  // Given an error thrown by an RPC call to a grain, return whether or not it makes sense to try                      // 113
  // to restart the grain and retry. `retryCount` is the number of times that the request has                          // 114
  // already gone through this cycle (should be zero for the first call).                                              // 115
                                                                                                                       // 116
  return error.kjType === "disconnected" && retryCount < 1;                                                            // 117
}                                                                                                                      // 118
                                                                                                                       // 119
SandstormBackend.prototype.maybeRetryUseGrain = function (grainId, cb, retryCount, err) {                              // 120
  var self = this;                                                                                                     // 121
  if (SandstormBackend.shouldRestartGrain(err, retryCount)) {                                                          // 122
    return inMeteor(function () {                                                                                      // 123
      return cb(self.openGrain(grainId, true).supervisor)                                                              // 124
          .catch(self.maybeRetryUseGrain.bind(undefined, grainId, cb, retryCount + 1));                                // 125
    });                                                                                                                // 126
  } else {                                                                                                             // 127
    throw err;                                                                                                         // 128
  }                                                                                                                    // 129
}                                                                                                                      // 130
                                                                                                                       // 131
SandstormBackend.prototype.useGrain = function (grainId, cb) {                                                         // 132
  // This will open a grain for you, handling restarts if needed, and call the passed function with                    // 133
  // the supervisor capability as the only parameter. The callback must return a promise that used                     // 134
  // the supervisor, so that we can check if a disconnect error occurred, and retry if possible.                       // 135
  // This function returns the same promise that your callback returns.                                                // 136
  //                                                                                                                   // 137
  // This function is NOT expected to be run in a meteor context.                                                      // 138
                                                                                                                       // 139
  var runningGrain = this.runningGrains[grainId];                                                                      // 140
  var self = this;                                                                                                     // 141
  if (runningGrain) {                                                                                                  // 142
    return runningGrain.then(function (grainInfo) {                                                                    // 143
      return cb(grainInfo.supervisor);                                                                                 // 144
    }).catch(self.maybeRetryUseGrain.bind(undefined, grainId, cb, 0));                                                 // 145
  } else {                                                                                                             // 146
    return inMeteor(function () {                                                                                      // 147
      return cb(self.openGrain(grainId, false).supervisor)                                                             // 148
          .catch(self.maybeRetryUseGrain.bind(undefined, grainId, cb, 0));                                             // 149
    });                                                                                                                // 150
  }                                                                                                                    // 151
}                                                                                                                      // 152
                                                                                                                       // 153
SandstormBackend.prototype.continueGrain = function(grainId) {                                                         // 154
  var grain = Grains.findOne(grainId);                                                                                 // 155
  if (!grain) {                                                                                                        // 156
    throw new Meteor.Error(404, "Grain Not Found", "Grain ID: " + grainId);                                            // 157
  }                                                                                                                    // 158
                                                                                                                       // 159
  // If a DevPackage with the same app ID is currently active, we let it override the installed                        // 160
  // package, so that the grain runs using the dev app.                                                                // 161
  var devPackage = DevPackages.findOne({_appId: grain.appId});                                                         // 162
  var isDev;                                                                                                           // 163
  if (devPackage) {                                                                                                    // 164
    isDev = true;                                                                                                      // 165
  } else {                                                                                                             // 166
    var pkg = Packages.findOne(grain.packageId);                                                                       // 167
    if (!pkg) {                                                                                                        // 168
      throw new Meteor.Error(500, "Grain's package not installed",                                                     // 169
                             "Package ID: " + grain.packageId);                                                        // 170
    }                                                                                                                  // 171
  }                                                                                                                    // 172
  var manifest = pkg.manifest;                                                                                         // 173
  var packageId = pkg._id;                                                                                             // 174
                                                                                                                       // 175
  if (!("continueCommand" in manifest)) {                                                                              // 176
    throw new Meteor.Error(500, "Package manifest defines no continueCommand.",                                        // 177
                           "Package ID: " + packageId);                                                                // 178
  }                                                                                                                    // 179
                                                                                                                       // 180
  return this.startGrainInternal(                                                                                      // 181
      packageId, grainId, grain.userId, manifest.continueCommand, false, isDev);                                       // 182
}                                                                                                                      // 183
                                                                                                                       // 184
SandstormBackend.prototype.startGrainInternal = function(packageId, grainId, ownerId, command, isNew, isDev) {         // 185
  // Starts the grain supervisor.  Must be executed in a Meteor context.  Blocks until grain is                        // 186
  // started. Returns a promise for an object containing two fields: `owner` (the ID of the owning                     // 187
  // user) and `supervisor` (the supervisor capability).                                                               // 188
                                                                                                                       // 189
  if (isUserExcessivelyOverQuota(Meteor.users.findOne(ownerId))) {                                                     // 190
    throw new Meteor.Error(402, "Cannot start grain because owner's storage is exhausted.");                           // 191
  }                                                                                                                    // 192
                                                                                                                       // 193
  // Ugly: Stay backwards-compatible with old manifests that had "executablePath" and "args" rather                    // 194
  //   than just "argv".                                                                                               // 195
  if ("args" in command) {                                                                                             // 196
    if (!("argv" in command)) {                                                                                        // 197
      command.argv = command.args;                                                                                     // 198
    }                                                                                                                  // 199
    delete command.args;                                                                                               // 200
  }                                                                                                                    // 201
  if ("executablePath" in command) {                                                                                   // 202
    if (!("deprecatedExecutablePath" in command)) {                                                                    // 203
      command.deprecatedExecutablePath = command.executablePath;                                                       // 204
    }                                                                                                                  // 205
    delete command.executablePath;                                                                                     // 206
  }                                                                                                                    // 207
                                                                                                                       // 208
  var whenReady = this._backendCap.startGrain(ownerId, grainId, packageId, command, isNew, isDev)                      // 209
      .then(function (results) {                                                                                       // 210
    return {                                                                                                           // 211
      owner: ownerId,                                                                                                  // 212
      supervisor: results.supervisor                                                                                   // 213
    };                                                                                                                 // 214
  });                                                                                                                  // 215
                                                                                                                       // 216
  this.runningGrains[grainId] = whenReady;                                                                             // 217
  return waitPromise(whenReady);                                                                                       // 218
}                                                                                                                      // 219
                                                                                                                       // 220
SandstormBackend.prototype.updateLastActive = function(grainId, userId, identityId) {                                  // 221
  // Update the lastActive date on the grain, any relevant API tokens, and the user,                                   // 222
  // and also update the user's storage usage.                                                                         // 223
                                                                                                                       // 224
  var storagePromise = undefined;                                                                                      // 225
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 226
    storagePromise = globalBackend._backendCap.getUserStorageUsage(userId);                                            // 227
  }                                                                                                                    // 228
                                                                                                                       // 229
  var now = new Date();                                                                                                // 230
  Grains.update(grainId, {$set: {lastUsed: now}});                                                                     // 231
  if (userId) {                                                                                                        // 232
    Meteor.users.update(userId, {$set: {lastActive: now}});                                                            // 233
  }                                                                                                                    // 234
  if (identityId) {                                                                                                    // 235
    // Update any API tokens that match this user/grain pairing as well                                                // 236
    var now = new Date();                                                                                              // 237
    ApiTokens.update({"grainId": grainId, "owner.user.identityId": identityId},                                        // 238
        {$set: {"owner.user.lastUsed": now }});                                                                        // 239
  }                                                                                                                    // 240
                                                                                                                       // 241
  if (Meteor.settings.public.quotaEnabled) {                                                                           // 242
    try {                                                                                                              // 243
      var ownerId = Grains.findOne(grainId).userId;                                                                    // 244
      var size = parseInt(waitPromise(storagePromise).size);                                                           // 245
      Meteor.users.update(ownerId, {$set: {storageUsage: size}});                                                      // 246
      // TODO(security): Consider actively killing grains if the user is excessively over quota?                       // 247
      //   Otherwise a constantly-active grain could consume arbitrary space without being stopped.                    // 248
    } catch (err) {                                                                                                    // 249
      if (err.kjType !== "unimplemented") {                                                                            // 250
        console.error("error getting user storage usage:", err.stack);                                                 // 251
      }                                                                                                                // 252
    }                                                                                                                  // 253
  }                                                                                                                    // 254
}                                                                                                                      // 255
                                                                                                                       // 256
                                                                                                                       // 257
function generateSessionId(grainId, userId, salt) {                                                                    // 258
  var sessionParts = [grainId, salt];                                                                                  // 259
  if (userId) {                                                                                                        // 260
    sessionParts.push(userId);                                                                                         // 261
  }                                                                                                                    // 262
  var sessionInput = sessionParts.join(":");                                                                           // 263
  return Crypto.createHash("sha256").update(sessionInput).digest("hex");                                               // 264
}                                                                                                                      // 265
                                                                                                                       // 266
SandstormBackend.prototype.openSessionInternal = function (grainId, userId, identityId, title, apiToken, cachedSalt) { // 267
                                                                                                                       // 268
  // Start the grain if it is not running. This is an optimization: if we didn't start it here,                        // 269
  // it would start on the first request to the session host, but we'd like to get started before                      // 270
  // the round trip.                                                                                                   // 271
  var runningGrain = this.runningGrains[grainId];                                                                      // 272
  var grainInfo;                                                                                                       // 273
  if (runningGrain) {                                                                                                  // 274
    grainInfo = waitPromise(runningGrain);                                                                             // 275
  } else {                                                                                                             // 276
    grainInfo = this.continueGrain(grainId);                                                                           // 277
  }                                                                                                                    // 278
                                                                                                                       // 279
  this.updateLastActive(grainId, userId, identityId);                                                                  // 280
                                                                                                                       // 281
  cachedSalt = cachedSalt || Random.id(22);                                                                            // 282
  var sessionId = generateSessionId(grainId, userId, cachedSalt);                                                      // 283
  var session = Sessions.findOne({_id: sessionId});                                                                    // 284
  if (session) {                                                                                                       // 285
    // TODO(someday): also do some more checks for anonymous sessions (sessions without a userId).                     // 286
    if ((session.identityId && session.identityId !== identityId) ||                                                   // 287
        (session.grainId !== grainId)) {                                                                               // 288
      var e = new Meteor.Error(500, "Duplicate SessionId");                                                            // 289
      console.error(e);                                                                                                // 290
      throw e;                                                                                                         // 291
    } else {                                                                                                           // 292
      return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};       // 293
    }                                                                                                                  // 294
  }                                                                                                                    // 295
                                                                                                                       // 296
  session = {                                                                                                          // 297
    _id: sessionId,                                                                                                    // 298
    grainId: grainId,                                                                                                  // 299
    hostId: Crypto.createHash("sha256").update(sessionId).digest("hex").slice(0, 32),                                  // 300
    timestamp: new Date().getTime(),                                                                                   // 301
    hasLoaded: false                                                                                                   // 302
  };                                                                                                                   // 303
                                                                                                                       // 304
  if (userId) {                                                                                                        // 305
    session.identityId = identityId;                                                                                   // 306
    session.userId = userId;                                                                                           // 307
  } else if (apiToken) {                                                                                               // 308
    session.hashedToken = apiToken._id;                                                                                // 309
  } else {                                                                                                             // 310
    // Must be old-style sharing, i.e. !grain.private.                                                                 // 311
  }                                                                                                                    // 312
                                                                                                                       // 313
  Sessions.insert(session);                                                                                            // 314
                                                                                                                       // 315
  return {sessionId: session._id, title: title, grainId: grainId, hostId: session.hostId, salt: cachedSalt};           // 316
}                                                                                                                      // 317
                                                                                                                       // 318
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-backend'] = {
  SandstormBackend: SandstormBackend
};

})();

//# sourceMappingURL=sandstorm-backend.js.map
